﻿using NetBootServerCore.Components.DDNS.Providers;
using NetBootServerCore.Components.Logging;

namespace NetBootServerCore.Components.DDNS
{
    public class DynamicDnsClient : NbsComponentBase
    {
        private static readonly IReadOnlyDictionary<string, DomainProvider> DomainProviders = new Dictionary<string, DomainProvider>
        {
            { "GoDaddy", new GoDaddyDomainProvider() },
            { "CloudFlare", new CloudFlareDomainProvider() }
        };

        private readonly IConfigurationSection _config;

        private bool _enable;
        private string _getMyIp;
        private DomainConfigEntry[] _domains;

        private bool _isRunning;
        private Task _task;

        public DynamicDnsClient(IConfigurationSection config) : base(new Logger("DDNS")) => _config = config;

        protected override void OnInit()
        {
            _enable = bool.Parse(_config["Enable"]);
            if (!_enable)
            {
                Logger.LogInfo("Disabled by config");
                return;
            }

            _getMyIp = _config["GetMyIp"];
            _domains = _config.GetSection("Domains").Get<DomainConfigEntry[]>().Where(p => p.Enable).ToArray();
        }

        protected override void OnStart()
        {
            if (!_enable)
            {
                Logger.LogInfo("Disabled by config");
                return;
            }

            string myIp, dnsIp;

            _isRunning = true;
            _task = Task.Run(() =>
            {
                while (_isRunning)
                {
                    try
                    {
                        Logger.LogVerbose("Getting my ip...");
                        myIp = GetMyIp();
                        Logger.LogVerbose($"My ip is: {myIp}");

                        foreach (var entry in _domains)
                        {
                            if (false == DomainProviders.TryGetValue(entry.Provider, out var provider))
                            {
                                Logger.LogWarn($"Not found domain provider:{entry.Provider}");
                                continue;
                            }

                            Logger.LogVerbose($"Getting dns ip for [{entry.RecordName} {entry.Domain} {entry.Provider}]...");
                            dnsIp = provider.GetIp(entry);
                            Logger.LogVerbose($"Dns ip is: {dnsIp??"(null)"}");

                            if (myIp != dnsIp)
                            {
                                Logger.LogInfo($"Updating dns record for [{entry.RecordName} {entry.Domain} {entry.Provider}]...");
                                provider.PutIp(entry, myIp);
                            }
                            else
                            {
                                Logger.LogVerbose("Nothing to do.");
                            }
                        }
                    }
                    catch (Exception e)
                    {
                        Logger.LogError("MainLoop", e);
                    }

                    Logger.LogVerbose("Sleeping...");
                    for (var i = 0; i < 600; i++)
                    {
                        if (_isRunning == false) break;
                        Thread.Sleep(1000);
                    }
                    Logger.LogVerbose("Wake up");
                }
            });
        }

        private string GetMyIp()
        {
            using var http = new HttpClient(new HttpClientHandler { UseProxy = false });
            var ret = http.GetAsync(_getMyIp).Result.Content.ReadAsStringAsync().Result;
            return ret;
        }

        protected override void OnStop()
        {
            if (!_enable)
            {
                Logger.LogInfo("Disabled by config");
                return;
            }
            _isRunning = false;
            _task.Wait(10 * 1000);
        }
    }
}
