﻿using Microsoft.AspNetCore.Http.Extensions;
using NetBootServerCore.Components.Logging;
using System.Text.RegularExpressions;

namespace NetBootServerCore.Components.Http.Processors
{
    public class BootScriptProcessor : NbsComponentBase, IHttpRequestProcessor
    {
        private readonly IConfigurationSection _config;
        private readonly Regex _regMac = new("^[0-9a-fA-F]{2}-[0-9a-fA-F]{2}-[0-9a-fA-F]{2}-[0-9a-fA-F]{2}-[0-9a-fA-F]{2}-[0-9a-fA-F]{2}$", RegexOptions.Compiled);

        private PathString _basePath;
        private string _rootDir;

        public BootScriptProcessor(IConfigurationSection config) : base(new Logger("HTTP.BSP")) => _config = config;

        protected override void OnInit()
        {
            _basePath = _config["Base"].TrimEnd('/');
            _rootDir = Path.GetFullPath(_config["RootDir"]);
        }

        public async Task<bool> ProcessRequest(HttpContext context)
        {
            if (!context.Request.Path.StartsWithSegments(_basePath, out var remaining)) return false;

            if (!remaining.StartsWithSegments("/ipxe", out remaining))
            {
                context.Response.StatusCode = 404;
                return true;
            }

            //TODO: auto match http prefix

            if (remaining == "")
            {
                // GET /boot/ipxe

                var prefix = new Uri(context.Request.GetDisplayUrl()).GetLeftPart(UriPartial.Authority);

                context.Response.StatusCode = 200;
                await context.Response.WriteAsync(
                    $"#!ipxe\r\n" +
                        $"chain {prefix}/boot/ipxe/script?mac=${{mac:hexhyp}}"
                );

                return true;
            }

            if (remaining.StartsWithSegments("/script"))
            {
                // GET /boot/ipxe/script?mac=11-22-33-44-55-66

                var displayUrl = context.Request.GetDisplayUrl();

                var mac = context.Request.Query["mac"].ToString().ToUpper();
                if (string.IsNullOrEmpty(mac) && false == _regMac.IsMatch(mac))
                {
                    return false;
                }

                var scriptPath = Path.Combine(_rootDir, $"iPXE-{mac}.txt");
                if (File.Exists(scriptPath))
                {
                    context.Response.StatusCode = 200;
                    await context.Response.SendFileAsync(scriptPath);
                    return true;
                }

                var defaultScriptPath = Path.Combine(_rootDir, "iPXE-Default.txt");
                if (File.Exists(defaultScriptPath))
                {
                    File.Copy(defaultScriptPath, scriptPath);
                    context.Response.StatusCode = 200;
                    await context.Response.SendFileAsync(defaultScriptPath);
                    return true;
                }

                context.Response.StatusCode = 206;
                return true;
            }

            return false;
        }
    }
}
