﻿using NetBootServerCore.Components.Logging;

namespace NetBootServerCore.Components.Http.Processors;

public class SanProcessor : NbsComponentBase, IHttpRequestProcessor
{
    private readonly IConfigurationSection _config;

    private PathString _basePath;
    private string _rootDir;

    public SanProcessor(IConfigurationSection config) : base(new Logger("HTTP.SAN")) => _config = config;

    protected override void OnInit()
    {
        _basePath = _config["Base"].TrimEnd('/');
        _rootDir = Path.GetFullPath(_config["RootDir"]);
    }

    public async Task<bool> ProcessRequest(HttpContext context)
    {
        if (!context.Request.Path.StartsWithSegments(_basePath, out var remaining)) return false;

        var path = Path.Combine(_rootDir, remaining.ToString().TrimStart('/'));
        var file = Path.GetFullPath(path);

        if (!file.StartsWith(_rootDir, StringComparison.InvariantCultureIgnoreCase))
        {
            Logger.LogWarn($"Denied invalid access {file}");
            return false;
        }

        if (!File.Exists(file)) return false;

        if (context.Request.Method == "HEAD")
        {
            await using var lenStream = File.OpenRead(file);
            context.Response.StatusCode = 204;
            context.Response.ContentLength = lenStream.Length;
            return true;
        }

        var x = context.Request.Headers.Range.ToString().Split(new[] { "=", "-" }, StringSplitOptions.RemoveEmptyEntries);
        if (x.Length == 0 || x.Length > 3)
        {
            context.Response.StatusCode = 200;
            await context.Response.SendFileAsync(file);
            return true;
        }

        context.Response.StatusCode = 206;
        var fs = File.OpenRead(file);

        if (x.Length == 2 && x[0] == "bytes" && int.TryParse(x[1], out var start))
        {
            Logger.LogDebug($"Request range: {start}- len:{fs.Length - start}");

            context.Response.Headers.Add("Content-Range", $"bytes {start}-{fs.Length - 1}/{fs.Length}");
            context.Response.ContentLength = fs.Length - start;
            await fs.CopyToAsync(context.Response.BodyWriter.AsStream());
            return true;
        }

        if (x.Length == 3 && x[0] == "bytes" && int.TryParse(x[1], out start) && int.TryParse(x[2], out var end))
        {
            var bodyLen = end - start + 1;
            Logger.LogDebug($"Request range: {start}-{end} len:{bodyLen}");

            context.Response.Headers.Add("Content-Range", $"bytes {end}-{start}/{fs.Length}");
            context.Response.ContentLength = bodyLen;

            const int megaBytes = 1024 * 1024;
            const int bufferChunk = megaBytes * 16;

            var outputStream = context.Response.BodyWriter.AsStream();

            var remainBytes = bodyLen;
            fs.Position = start;
            while (remainBytes > 0)
            {
                var buf = new byte[remainBytes > bufferChunk ? bufferChunk : remainBytes];

                var r = fs.Read(buf, 0, buf.Length);
                await outputStream.WriteAsync(buf, 0, r);

                remainBytes -= r;
            }
            return true;
        }

        return false;
    }
}
