using NetBootServerCore.Components.DDNS;
using NetBootServerCore.Components.Dns;
using NetBootServerCore.Components.Http;
using NetBootServerCore.Components.Http.Processors;
using NetBootServerCore.Components.Http.Processors.WebApi;
using NetBootServerCore.Components.LivingConsole;
using NetBootServerCore.Components.Logging;
using NetBootServerCore.Components.NetBoot;
using NetBootServerCore.Components.NetBoot.Dhcp;

namespace NetBootServerCore;

public static class NbsProgram
{
    public static readonly DateTime LaunchTime;

    static NbsProgram()
    {
        LaunchTime = DateTime.Now;
    }

    private static void Main(string[] args)
    {
        LiveConsole.Start();

        var logger = new Logger("APP");

        logger.LogInfo($"Launch Time: {LaunchTime:yyyy-MM-dd HH:mm:ss.fff}");

        logger.LogDebug("Logger Init");
        logger.LogVerbose("Logger Init");
        logger.LogInfo("Logger Init");
        logger.LogWarn("Logger Init");
        logger.LogError("Logger Init");
        logger.LogFatal("Logger Init");

        logger.LogInfo("Load config");
        Environment.CurrentDirectory = AppDomain.CurrentDomain.BaseDirectory;
        var rootConfig = WebApplication.CreateBuilder(args).Configuration;
        var httpConfig = rootConfig.GetSection("HTTP");
        var httpPrefix = httpConfig["Prefix"];
        Console.Title = $"{nameof(NetBootServerCore)} at {httpPrefix}";
        Environment.SetEnvironmentVariable("ASPNETCORE_URLS", httpPrefix);

        logger.LogInfo("Create components");
        var dhcp = new DhcpServer(rootConfig.GetSection("DHCP"));
        var tftp = new NbsTftpServer(rootConfig.GetSection("TFTP"));
        var dnsForwarder = new DnsForwarder(rootConfig.GetSection("DNSF"));
        var ddns = new DynamicDnsClient(rootConfig.GetSection("DDNS"));

        var bootScriptProcessor = new BootScriptProcessor(httpConfig.GetSection("BootScript"));
        var sanProcessor = new SanProcessor(httpConfig.GetSection("SAN"));
        var uiProcessor = new WebUIProcessor(httpConfig.GetSection("WebUI"));
        var apiProcessor = new WebApiProcessor(httpConfig.GetSection("WebApi"), dhcp, uiProcessor);
        var httpServer = new HttpServer(args, new IHttpRequestProcessor[]
        {
            bootScriptProcessor,
            sanProcessor,
            apiProcessor,
            uiProcessor
        });

        logger.LogInfo("Init components");
        dhcp.Init();
        tftp.Init();
        dnsForwarder.Init();
        ddns.Init();
        bootScriptProcessor.Init();
        sanProcessor.Init();
        apiProcessor.Init();
        uiProcessor.Init();
        httpServer.Init();

        logger.LogInfo("Start components");
        dhcp.Start();
        tftp.Start();
        dnsForwarder.Start();
        ddns.Start();
        bootScriptProcessor.Start();
        sanProcessor.Start();
        httpServer.Start();
        apiProcessor.Start();
        uiProcessor.Start();

        logger.LogInfo("Running");

        //Waiting until Control-C
        {
            CancellationTokenSource ctSource = new();
            var ct = ctSource.Token;

            void ExitHandler()
            {
                // You can add any arbitrary global clean up
                logger.LogInfo("Exiting...");
                ctSource.Cancel();
            }

            // Assign exit handler to be called when the process is terminated
            // or the user hits CTRL+C
            AppDomain.CurrentDomain.ProcessExit += (_, _) => ExitHandler();
            Console.CancelKeyPress += (_, _) => ExitHandler();

            // Then you can use the cancellation token to check for exit:
            while (!ct.IsCancellationRequested)
            {
                try
                {
                    Task.Delay(1000, ct).Wait(ct);
                }
                catch (OperationCanceledException e)
                {
                    break;
                }
            }
        }

        //stop components
        dhcp.Stop();
        tftp.Stop();
        dnsForwarder.Stop();
        ddns.Stop();
        uiProcessor.Stop();
        apiProcessor.Stop();
        sanProcessor.Stop();
        bootScriptProcessor.Stop();
        httpServer.Stop();

        logger.LogInfo("Finished");

        LiveConsole.Stop();
    }
}
